from czsc.data.jq import get_kline
from czsc.data import freq_map
from czsc.signals import KlineSignals

def test_signals():
    kline = get_kline(symbol="300033.XSHE", end_date="20201128", freq="5min", count=1000)
    ks = KlineSignals(kline, name=freq_map.get("5min", "本级别"), bi_mode="new", max_count=2000)
    print(ks.get_signals())





# coding: utf-8
from datetime import datetime
import traceback
from pyecharts.charts import Tab
from pyecharts.components import Table
from pyecharts.options import ComponentTitleOpts

from .signals import KlineSignals
from .data.jq import get_kline
from .data import freq_map
from .utils.plot import ka_to_echarts


class CzscTrader:
    """缠中说禅股票 选股/择时"""
    def __init__(self, symbol):
        """
        :param symbol:
        """
        self.symbol = symbol
        self.__generate_signals()
        self.freqs = ['1分钟', '5分钟', '30分钟', '日线']

    def __generate_signals(self):
        self.signals = {"symbol": self.symbol}
        self.kas = dict()
        for freq in ['1min', '5min', '30min', 'D']:
            try:
                kline = get_kline(symbol=self.symbol, end_date=datetime.now(), freq=freq, count=300)
                ks = KlineSignals(kline, name=freq_map.get(freq, "本级别"), bi_mode="new", max_count=300, use_xd=False)
                self.signals.update(ks.get_signals())
                self.kas[freq_map.get(freq, "本级别")] = ks
            except:
                traceback.print_exc()
        self.end_dt = self.kas['1分钟'].end_dt
        self.latest_price = self.kas['1分钟'].kline_raw[-1]['close']

    def run_selector(self):
        """执行选股：优先输出大级别的机会"""
        s = self.signals
        if s['30分钟_第N笔涨跌力度'] == '向下笔新低盘背' or s['5分钟_五笔趋势类背驰'] == 'down':
            if s['日线_三笔回调构成第三买卖点'] == '三买':
                return "日线三笔回调构成第三买点"

            if s["日线_第N笔第三买卖"] == '三买':
                return "日线第三买点"

            if s['日线_五笔趋势类背驰'] == 'down':
                return "日线五笔下跌趋势类背驰"

            if s['日线_第N笔出井'] == "向下小井" and s['日线_第N-2笔出井'] == "向下小井":
                return "日线连续两次向下小井"

            if s['日线_第N笔出井'] == "向下大井":
                return "日线向下大井"

        if s['5分钟_第N笔涨跌力度'] == '向下笔新低盘背' or s['1分钟_五笔趋势类背驰'] == 'down':
            if s['30分钟_三笔回调构成第三买卖点'] == '三买':
                return "30分钟三笔回调构成第三买点"

            if s["30分钟_第N笔第三买卖"] == '三买':
                return "30分钟第三买点"

            if s['30分钟_五笔趋势类背驰'] == 'down':
                return "30分钟五笔下跌趋势类背驰"
        return "other"

    def take_snapshot(self, file_html, width="950px", height="480px"):
        tab = Tab(page_title="{}的交易快照@{}".format(self.symbol, self.end_dt.strftime("%Y-%m-%d %H:%M")))
        for freq in self.freqs:
            chart = ka_to_echarts(self.kas[freq], width, height)
            tab.add(chart, freq)

        headers = ["名称", "数据"]
        rows = [[k, v] for k, v in self.signals.items()]
        table = Table()
        table.add(headers, rows)
        table.set_global_opts(title_opts=ComponentTitleOpts(title="缠论信号", subtitle=""))
        tab.add(table, "信号表")
        tab.render(file_html)

    def long_open(self):
        """执行开仓择时"""
        s = self.signals

    def long_close(self):
        """执行平仓择时"""
        s = self.signals



# coding: utf-8
import os
import pickle
import json
import requests
import warnings
import pandas as pd
from datetime import datetime, timedelta

url = "https://dataapi.joinquant.com/apis"
home_path = os.path.expanduser("~")
file_token = os.path.join(home_path, "jq.token")

def set_token(jq_mob, jq_pwd):
    """

    :param jq_mob: str
        mob是申请JQData时所填写的手机号
    :param jq_pwd: str
        Password为聚宽官网登录密码，新申请用户默认为手机号后6位
    :return: None
    """
    with open(file_token, 'wb') as f:
        pickle.dump([jq_mob, jq_pwd], f)


def get_token():
    """获取调用凭证"""
    if not os.path.exists(file_token):
        raise ValueError(f"{file_token} 文件不存在，请先调用 set_token 进行设置")

    with open(file_token, 'rb') as f:
        jq_mob, jq_pwd = pickle.load(f)

    body = {
        "method": "get_current_token",
        "mob": jq_mob,  # mob是申请JQData时所填写的手机号
        "pwd": jq_pwd,  # Password为聚宽官网登录密码，新申请用户默认为手机号后6位
    }
    response = requests.post(url, data=json.dumps(body))
    token = response.text
    return token


def text2df(text):
    rows = [x.split(",") for x in text.strip().split('\n')]
    df = pd.DataFrame(rows[1:], columns=rows[0])
    return df


def get_query_count() -> int:
    """获取查询剩余条数
    https://dataapi.joinquant.com/docs#get_query_count---%E8%8E%B7%E5%8F%96%E6%9F%A5%E8%AF%A2%E5%89%A9%E4%BD%99%E6%9D%A1%E6%95%B0

    :return: int
    """
    data = {
        "method": "get_query_count",
        "token": get_token(),
    }
    r = requests.post(url, data=json.dumps(data))
    return int(r.text)


def get_concepts():
    """获取概念列表

    https://dataapi.joinquant.com/docs#get_concepts---%E8%8E%B7%E5%8F%96%E6%A6%82%E5%BF%B5%E5%88%97%E8%A1%A8

    :return: df
    """
    data = {
        "method": "get_concepts",
        "token": get_token(),
    }
    r = requests.post(url, data=json.dumps(data))
    df = text2df(r.text)
    return df

def get_concept_stocks(symbol, date=None):
    """获取概念成份股

    https://dataapi.joinquant.com/docs#get_concept_stocks---%E8%8E%B7%E5%8F%96%E6%A6%82%E5%BF%B5%E6%88%90%E4%BB%BD%E8%82%A1

    :param symbol: str
        如 GN036
    :param date: str or datetime
        日期，如 2020-08-08
    :return: list

    examples:
    -------
    >>> symbols1 = get_concept_stocks("GN036", date="2020-07-08")
    >>> symbols2 = get_concept_stocks("GN036", date=datetime.now())
    """
    if not date:
        date = str(datetime.now().date())

    if isinstance(date, datetime):
        date = str(date.date())

    data = {
        "method": "get_concept_stocks",
        "token": get_token(),
        "code": symbol,
        "date": date
    }
    r = requests.post(url, data=json.dumps(data))
    return r.text.split('\n')

def get_index_stocks(symbol, date=None):
    """获取指数成份股

    https://dataapi.joinquant.com/docs#get_index_stocks---%E8%8E%B7%E5%8F%96%E6%8C%87%E6%95%B0%E6%88%90%E4%BB%BD%E8%82%A1

    :param symbol: str
        如 000300.XSHG
    :param date: str or datetime
        日期，如 2020-08-08
    :return: list

    examples:
    -------
    >>> symbols1 = get_index_stocks("000300.XSHG", date="2020-07-08")
    >>> symbols2 = get_index_stocks("000300.XSHG", date=datetime.now())
    """
    if not date:
        date = str(datetime.now().date())

    if isinstance(date, datetime):
        date = str(date.date())

    data = {
        "method": "get_index_stocks",
        "token": get_token(),
        "code": symbol,
        "date": date
    }
    r = requests.post(url, data=json.dumps(data))
    return r.text.split('\n')

def get_all_securities(code, date=None) -> pd.DataFrame:
    """
    https://dataapi.joinquant.com/docs#get_all_securities---%E8%8E%B7%E5%8F%96%E6%89%80%E6%9C%89%E6%A0%87%E7%9A%84%E4%BF%A1%E6%81%AF
    获取平台支持的所有股票、基金、指数、期货信息

    参数：

    code: 证券类型,可选: stock, fund, index, futures, etf, lof, fja, fjb, QDII_fund,
                        open_fund, bond_fund, stock_fund, money_market_fund, mixture_fund, options
    date: 日期，用于获取某日期还在上市的证券信息，date为空时表示获取所有日期的标的信息

    :return:
    """
    if not date:
        date = str(datetime.now().date())

    if isinstance(date, datetime):
        date = str(date.date())

    data = {
        "method": "get_all_securities",
        "token": get_token(),
        "code": code,
        "date": date
    }
    r = requests.post(url, data=json.dumps(data))
    return text2df(r.text)

def get_kline(symbol,  end_date, freq, start_date=None, count=None):
    """获取K线数据

    :param symbol: str
        聚宽标的代码
    :param start_date: datetime
        截止日期
    :param end_date: datetime
        截止日期
    :param freq: str
        K线级别，可选值 ['1min', '5min', '30min', '60min', 'D', 'W', 'M']
    :param count: int
        K线数量，最大值为 5000
    :return: pd.DataFrame

    >>> start_date = datetime.strptime("20200701", "%Y%m%d")
    >>> end_date = datetime.strptime("20200719", "%Y%m%d")
    >>> df1 = get_kline(symbol="000001.XSHG", start_date=start_date, end_date=end_date, freq="1min")
    >>> df2 = get_kline(symbol="000001.XSHG", end_date=end_date, freq="1min", count=1000)
    >>> df3 = get_kline(symbol="000001.XSHG", start_date='20200701', end_date='20200719', freq="1min")
    >>> df4 = get_kline(symbol="000001.XSHG", end_date='20200719', freq="1min", count=1000)
    """
    if count and count > 5000:
        warnings.warn(f"count={count}, 超过5000的最大值限制，仅返回最后5000条记录")

    # 1m, 5m, 15m, 30m, 60m, 120m, 1d, 1w, 1M
    freq_convert = {"1min": "1m", "5min": '5m', '15min': '15m',
                    "30min": "30m", "60min": '60m', "D": "1d", "W": '1w', "M": "1M"}
    end_date = pd.to_datetime(end_date)
    if start_date:
        start_date = pd.to_datetime(start_date)
        data = {
            "method": "get_price_period",
            "token": get_token(),
            "code": symbol,
            "unit": freq_convert[freq],
            "date": start_date.strftime("%Y-%m-%d"),
            "end_date": end_date.strftime("%Y-%m-%d"),
            # "fq_ref_date": end_date
        }
    elif count:
        data = {
            "method": "get_price",
            "token": get_token(),
            "code": symbol,
            "count": count,
            "unit": freq_convert[freq],
            "end_date": end_date.strftime("%Y-%m-%d"),
            # "fq_ref_date": end_date
        }
    else:
        raise ValueError("start_date 和 count 不能同时为空")

    r = requests.post(url, data=json.dumps(data))
    df = text2df(r.text)
    df['symbol'] = symbol
    df.rename({'date': 'dt', 'volume': 'vol'}, axis=1, inplace=True)
    df = df[['symbol', 'dt', 'open', 'close', 'high', 'low', 'vol']]
    for col in ['open', 'close', 'high', 'low', 'vol']:
        df.loc[:, col] = df[col].apply(lambda x: round(float(x), 2))
    df.loc[:, "dt"] = pd.to_datetime(df['dt'])
    return df

def download_kline(symbol, freq, start_date, end_date, delta, save=True):
    """下载K线数据

    :param save:
    :param symbol:
    :param end_date:
    :param freq:
    :param start_date:
    :param delta:
    :return:

    >>> start_date = datetime.strptime("20200101", "%Y%m%d")
    >>> end_date = datetime.strptime("20200719", "%Y%m%d")
    >>> df = download_kline("000001.XSHG", "1min", start_date, end_date, delta=timedelta(days=10), save=False)
    """
    start_date = pd.to_datetime(start_date)
    end_date = pd.to_datetime(end_date)

    data = []
    end_dt = start_date + delta
    print("开始下载数据：{} - {} - {}".format(symbol, start_date, end_date))
    df_ = get_kline(symbol, start_date=start_date, end_date=end_dt, freq=freq)
    if not df_.empty:
        data.append(df_)

    while end_dt < end_date:
        df_ = get_kline(symbol, start_date=start_date, end_date=end_dt, freq=freq)
        if not df_.empty:
            data.append(df_)
        start_date = end_dt
        end_dt += delta
        print("当前下载进度：{} - {} - {}".format(symbol, start_date, end_dt))

    df = pd.concat(data, ignore_index=True)
    print("{} 去重前K线数量为 {}".format(symbol, len(df)))
    df.drop_duplicates(['dt'], inplace=True)
    df.sort_values('dt', ascending=True, inplace=True)
    df.reset_index(drop=True, inplace=True)
    print("{} 去重后K线数量为 {}".format(symbol, len(df)))

    if save:
        df.to_csv(f"{symbol}_{freq}_{start_date.date()}_{end_date.date()}.csv", index=False, encoding="utf-8")
    else:
        return df

def get_fundamental(table: str, symbol: str, date: str, columns: str = "") -> dict:
    """
    https://dataapi.joinquant.com/docs#get_fundamentals---%E8%8E%B7%E5%8F%96%E5%9F%BA%E6%9C%AC%E8%B4%A2%E5%8A%A1%E6%95%B0%E6%8D%AE

    财务数据列表：
    https://www.joinquant.com/help/api/help?name=Stock#%E8%B4%A2%E5%8A%A1%E6%95%B0%E6%8D%AE%E5%88%97%E8%A1%A8

    :param table:
    :param symbol:
    :param date: str
        查询日期2019-03-04或者年度2018或者季度2018q1 2018q2 2018q3 2018q4
    :param columns:
    :return: df

    example:
    ============
    >>>> x1 = get_fundamental(table="indicator", symbol="300803.XSHE", date="2020-11-12")
    >>>> x2 = get_fundamental(table="indicator", symbol="300803.XSHE", date="2020")
    >>>> x3 = get_fundamental(table="indicator", symbol="300803.XSHE", date="2020q3")
    """
    data = {
        "method": "get_fundamentals",
        "token": get_token(),
        "table": table,
        "columns": columns,
        "code": symbol,
        "date": date,
        "count": 1
    }
    r = requests.post(url, data=json.dumps(data))
    df = text2df(r.text)
    try:
        return df.iloc[0].to_dict()
    except:
        return {}







# coding: utf-8

import warnings

try:
    import talib as ta
except ImportError:
    ta_lib_hint = "没有安装 ta-lib !!! 请到 https://www.lfd.uci.edu/~gohlke/pythonlibs/#ta-lib " \
                  "下载对应版本安装，预计分析速度提升2倍"
    warnings.warn(ta_lib_hint)
    from .utils import ta
import pandas as pd
import numpy as np
from .utils.plot import ka_to_image


def find_zs(points):
    """输入笔或线段标记点，输出中枢识别结果"""
    if len(points) < 5:
        return []

    # 当输入为笔的标记点时，新增 xd 值
    for j, x in enumerate(points):
        if x.get("bi", 0):
            points[j]['xd'] = x["bi"]

    def __get_zn(zn_points_):
        """把与中枢方向一致的次级别走势类型称为Z走势段，按中枢中的时间顺序，
        分别记为Zn等，而相应的高、低点分别记为gn、dn"""
        if len(zn_points_) % 2 != 0:
            zn_points_ = zn_points_[:-1]

        if zn_points_[0]['fx_mark'] == "d":
            z_direction = "up"
        else:
            z_direction = "down"

        zn = []
        for i in range(0, len(zn_points_), 2):
            zn_ = {
                "start_dt": zn_points_[i]['dt'],
                "end_dt": zn_points_[i + 1]['dt'],
                "high": max(zn_points_[i]['xd'], zn_points_[i + 1]['xd']),
                "low": min(zn_points_[i]['xd'], zn_points_[i + 1]['xd']),
                "direction": z_direction
            }
            zn_['mid'] = zn_['low'] + (zn_['high'] - zn_['low']) / 2
            zn.append(zn_)
        return zn

    k_xd = points
    k_zs = []
    zs_xd = []

    for i in range(len(k_xd)):
        if len(zs_xd) < 5:
            zs_xd.append(k_xd[i])
            continue
        xd_p = k_xd[i]
        zs_d = max([x['xd'] for x in zs_xd[:4] if x['fx_mark'] == 'd'])
        zs_g = min([x['xd'] for x in zs_xd[:4] if x['fx_mark'] == 'g'])
        if zs_g <= zs_d:
            zs_xd.append(k_xd[i])
            zs_xd.pop(0)
            continue

        # 定义四个指标,GG=max(gn),G=min(gn),D=max(dn),DD=min(dn)，n遍历中枢中所有Zn。
        # 定义ZG=min(g1、g2), ZD=max(d1、d2)，显然，[ZD，ZG]就是缠中说禅走势中枢的区间
        if xd_p['fx_mark'] == "d" and xd_p['xd'] > zs_g:
            zn_points = zs_xd[3:]
            # 线段在中枢上方结束，形成三买
            k_zs.append({
                'ZD': zs_d,
                "ZG": zs_g,
                'G': min([x['xd'] for x in zs_xd if x['fx_mark'] == 'g']),
                'GG': max([x['xd'] for x in zs_xd if x['fx_mark'] == 'g']),
                'D': max([x['xd'] for x in zs_xd if x['fx_mark'] == 'd']),
                'DD': min([x['xd'] for x in zs_xd if x['fx_mark'] == 'd']),
                'start_point': zs_xd[1],
                'end_point': zs_xd[-2],
                "zn": __get_zn(zn_points),
                "points": zs_xd,
                "third_buy": xd_p
            })
            zs_xd = []
        elif xd_p['fx_mark'] == "g" and xd_p['xd'] < zs_d:
            zn_points = zs_xd[3:]
            # 线段在中枢下方结束，形成三卖
            k_zs.append({
                'ZD': zs_d,
                "ZG": zs_g,
                'G': min([x['xd'] for x in zs_xd if x['fx_mark'] == 'g']),
                'GG': max([x['xd'] for x in zs_xd if x['fx_mark'] == 'g']),
                'D': max([x['xd'] for x in zs_xd if x['fx_mark'] == 'd']),
                'DD': min([x['xd'] for x in zs_xd if x['fx_mark'] == 'd']),
                'start_point': zs_xd[1],
                'end_point': zs_xd[-2],
                "points": zs_xd,
                "zn": __get_zn(zn_points),
                "third_sell": xd_p
            })
            zs_xd = []
        else:
            zs_xd.append(xd_p)

    if len(zs_xd) >= 5:
        zs_d = max([x['xd'] for x in zs_xd[:4] if x['fx_mark'] == 'd'])
        zs_g = min([x['xd'] for x in zs_xd[:4] if x['fx_mark'] == 'g'])
        if zs_g > zs_d:
            zn_points = zs_xd[3:]
            k_zs.append({
                'ZD': zs_d,
                "ZG": zs_g,
                'G': min([x['xd'] for x in zs_xd if x['fx_mark'] == 'g']),
                'GG': max([x['xd'] for x in zs_xd if x['fx_mark'] == 'g']),
                'D': max([x['xd'] for x in zs_xd if x['fx_mark'] == 'd']),
                'DD': min([x['xd'] for x in zs_xd if x['fx_mark'] == 'd']),
                'start_point': zs_xd[1],
                'end_point': None,
                "zn": __get_zn(zn_points),
                "points": zs_xd,
            })
    return k_zs


def has_gap(k1, k2, min_gap=0.002):
    """判断 k1, k2 之间是否有缺口"""
    assert k2['dt'] > k1['dt']
    if k1['high'] < k2['low'] * (1 - min_gap) \
            or k2['high'] < k1['low'] * (1 - min_gap):
        return True
    else:
        return False


def get_potential_xd(bi_points):
    """获取潜在线段标记点

    :param bi_points: list of dict
        笔标记点
    :return: list of dict
        潜在线段标记点
    """
    xd_p = []
    bi_d = [x for x in bi_points if x['fx_mark'] == 'd']
    bi_g = [x for x in bi_points if x['fx_mark'] == 'g']
    for i in range(1, len(bi_d) - 1):
        d1, d2, d3 = bi_d[i - 1: i + 2]
        if d1['bi'] > d2['bi'] < d3['bi']:
            xd_p.append(d2)
    for j in range(1, len(bi_g) - 1):
        g1, g2, g3 = bi_g[j - 1: j + 2]
        if g1['bi'] < g2['bi'] > g3['bi']:
            xd_p.append(g2)

    xd_p = sorted(xd_p, key=lambda x: x['dt'], reverse=False)
    return xd_p


class KlineAnalyze:
    def __init__(self, kline, name="本级别", bi_mode="new", max_count=1000,
                 use_xd=False, use_ta=True, ma_params=(5, 34, 120), verbose=False):
        """

        :param kline: list or pd.DataFrame
        :param name: str
        :param bi_mode: str
            new 新笔；old 老笔；默认值为 new
        :param max_count: int
            最大保存的K线数量
        :param use_xd: bool
            是否进行线段识别，对于以笔作为 f0 的交易策略而言，不进行线段识别可以加快分析速度
        :param use_ta: bool
            是否进行辅助技术指标的计算，不进行辅助技术指标的计算可以加快分析速度
        :param ma_params: tuple of int
            均线系统参数
        :param verbose: bool
        """
        self.name = name
        self.verbose = verbose
        self.bi_mode = bi_mode
        self.max_count = max_count
        self.use_xd = use_xd
        self.use_ta = use_ta
        self.ma_params = ma_params
        self.kline_raw = []  # 原始K线序列
        self.kline_new = []  # 去除包含关系的K线序列

        # 辅助技术指标
        self.ma = []
        self.macd = []

        # 分型、笔、线段
        self.fx_list = []
        self.bi_list = []
        self.xd_list = []

        # 根据输入K线初始化
        if isinstance(kline, pd.DataFrame):
            columns = kline.columns.to_list()
            self.kline_raw = [{k: v for k, v in zip(columns, row)} for row in kline.values]
        else:
            self.kline_raw = kline

        self.symbol = self.kline_raw[0]['symbol']
        self.start_dt = self.kline_raw[0]['dt']
        self.end_dt = self.kline_raw[-1]['dt']
        self.latest_price = self.kline_raw[-1]['close']

        self._update_kline_new()

        if self.use_ta:
            self._update_ta()

        self._update_fx_list()
        self._update_bi_list()

        if self.use_xd:
            self._update_xd_list()

    def _update_ta(self):
        """更新辅助技术指标"""
        if not self.ma:
            ma_temp = dict()
            close_ = np.array([x["close"] for x in self.kline_raw], dtype=np.double)
            for p in self.ma_params:
                ma_temp['ma%i' % p] = ta.SMA(close_, p)

            for i in range(len(self.kline_raw)):
                ma_ = {'ma%i' % p: ma_temp['ma%i' % p][i] for p in self.ma_params}
                ma_.update({"dt": self.kline_raw[i]['dt']})
                self.ma.append(ma_)
        else:
            ma_ = {'ma%i' % p: sum([x['close'] for x in self.kline_raw[-p:]]) / p
                   for p in self.ma_params}
            ma_.update({"dt": self.kline_raw[-1]['dt']})
            if self.verbose:
                print("ma new: %s" % str(ma_))

            if self.kline_raw[-2]['dt'] == self.ma[-1]['dt']:
                self.ma.append(ma_)
            else:
                self.ma[-1] = ma_

        assert self.ma[-2]['dt'] == self.kline_raw[-2]['dt']

        if not self.macd:
            close_ = np.array([x["close"] for x in self.kline_raw], dtype=np.double)
            # m1 is diff; m2 is dea; m3 is macd
            m1, m2, m3 = ta.MACD(close_, fastperiod=12, slowperiod=26, signalperiod=9)
            for i in range(len(self.kline_raw)):
                self.macd.append({
                    "dt": self.kline_raw[i]['dt'],
                    "diff": m1[i],
                    "dea": m2[i],
                    "macd": m3[i]
                })
        else:
            close_ = np.array([x["close"] for x in self.kline_raw[-200:]], dtype=np.double)
            # m1 is diff; m2 is dea; m3 is macd
            m1, m2, m3 = ta.MACD(close_, fastperiod=12, slowperiod=26, signalperiod=9)
            macd_ = {
                "dt": self.kline_raw[-1]['dt'],
                "diff": m1[-1],
                "dea": m2[-1],
                "macd": m3[-1]
            }
            if self.verbose:
                print("macd new: %s" % str(macd_))

            if self.kline_raw[-2]['dt'] == self.macd[-1]['dt']:
                self.macd.append(macd_)
            else:
                self.macd[-1] = macd_

        assert self.macd[-2]['dt'] == self.kline_raw[-2]['dt']

    def _update_kline_new(self):
        """更新去除包含关系的K线序列"""
        if len(self.kline_new) < 4:
            for x in self.kline_raw[:4]:
                self.kline_new.append(dict(x))

        # 新K线只会对最后一个去除包含关系K线的结果产生影响
        self.kline_new = self.kline_new[:-2]
        if len(self.kline_new) <= 4:
            right_k = [x for x in self.kline_raw if x['dt'] > self.kline_new[-1]['dt']]
        else:
            right_k = [x for x in self.kline_raw[-100:] if x['dt'] > self.kline_new[-1]['dt']]

        if len(right_k) == 0:
            return

        for k in right_k:
            k = dict(k)
            last_kn = self.kline_new[-1]
            if self.kline_new[-1]['high'] > self.kline_new[-2]['high']:
                direction = "up"
            else:
                direction = "down"

            # 判断是否存在包含关系
            cur_h, cur_l = k['high'], k['low']
            last_h, last_l = last_kn['high'], last_kn['low']
            if (cur_h <= last_h and cur_l >= last_l) or (cur_h >= last_h and cur_l <= last_l):
                self.kline_new.pop(-1)
                # 有包含关系，按方向分别处理
                if direction == "up":
                    last_h = max(last_h, cur_h)
                    last_l = max(last_l, cur_l)
                elif direction == "down":
                    last_h = min(last_h, cur_h)
                    last_l = min(last_l, cur_l)
                else:
                    raise ValueError

                k.update({"high": last_h, "low": last_l})
                # 保留红绿不变
                if k['open'] >= k['close']:
                    k.update({"open": last_h, "close": last_l})
                else:
                    k.update({"open": last_l, "close": last_h})
            self.kline_new.append(k)

    def _update_fx_list(self):
        """更新分型序列

        分型标记对象样例：
         {'dt': Timestamp('2020-11-26 00:00:00'),
          'fx_mark': 'd',
          'fx': 138.0,
          'start_dt': Timestamp('2020-11-25 00:00:00'),
          'end_dt': Timestamp('2020-11-27 00:00:00'),
          'fx_high': 144.87,
          'fx_low': 138.0}

         {'dt': Timestamp('2020-12-02 00:00:00'),
          'fx_mark': 'g',
          'fx': 150.67,
          'start_dt': Timestamp('2020-12-01 00:00:00'),
          'end_dt': Timestamp('2020-12-03 00:00:00'),
          'fx_high': 150.67,
          'fx_low': 141.6}
        """
        if len(self.kline_new) < 3:
            return

        self.fx_list = self.fx_list[:-1]
        if len(self.fx_list) == 0:
            kn = self.kline_new
        else:
            kn = [x for x in self.kline_new[-100:] if x['dt'] >= self.fx_list[-1]['dt']]

        i = 1
        while i <= len(kn) - 2:
            k1, k2, k3 = kn[i - 1: i + 2]

            if k1['high'] < k2['high'] > k3['high']:
                if self.verbose:
                    print("顶分型：{} - {} - {}".format(k1['dt'], k2['dt'], k3['dt']))
                fx = {
                    "dt": k2['dt'],
                    "fx_mark": "g",
                    "fx": k2['high'],
                    "start_dt": k1['dt'],
                    "end_dt": k3['dt'],
                    "fx_high": k2['high'],
                    "fx_low": k2['low'] if has_gap(k1, k2) else k1['low'],
                }
                self.fx_list.append(fx)

            elif k1['low'] > k2['low'] < k3['low']:
                if self.verbose:
                    print("底分型：{} - {} - {}".format(k1['dt'], k2['dt'], k3['dt']))
                fx = {
                    "dt": k2['dt'],
                    "fx_mark": "d",
                    "fx": k2['low'],
                    "start_dt": k1['dt'],
                    "end_dt": k3['dt'],
                    "fx_high": k2['high'] if has_gap(k1, k2) else k1['high'],
                    "fx_low": k2['low'],
                }
                self.fx_list.append(fx)

            else:
                if self.verbose:
                    print("无分型：{} - {} - {}".format(k1['dt'], k2['dt'], k3['dt']))
            i += 1

    def _update_bi_list(self):
        """更新笔序列

        笔标记对象样例：
         {'dt': Timestamp('2020-11-26 00:00:00'),
          'fx_mark': 'd',
          'start_dt': Timestamp('2020-11-25 00:00:00'),
          'end_dt': Timestamp('2020-11-27 00:00:00'),
          'fx_high': 144.87,
          'fx_low': 138.0,
          'bi': 138.0}

         {'dt': Timestamp('2020-12-02 00:00:00'),
          'fx_mark': 'g',
          'start_dt': Timestamp('2020-12-01 00:00:00'),
          'end_dt': Timestamp('2020-12-03 00:00:00'),
          'fx_high': 150.67,
          'fx_low': 141.6,
          'bi': 150.67}
        """
        if len(self.fx_list) < 2:
            return

        self.bi_list = self.bi_list[:-2]
        if len(self.bi_list) == 0:
            for fx in self.fx_list[:2]:
                bi = dict(fx)
                bi['bi'] = bi.pop('fx')
                self.bi_list.append(bi)

        if len(self.bi_list) <= 2:
            right_fx = [x for x in self.fx_list if x['dt'] > self.bi_list[-1]['dt']]
            if self.bi_mode == "old":
                right_kn = [x for x in self.kline_new if x['dt'] >= self.bi_list[-1]['dt']]
            elif self.bi_mode == 'new':
                right_kn = [x for x in self.kline_raw if x['dt'] >= self.bi_list[-1]['dt']]
            else:
                raise ValueError
        else:
            right_fx = [x for x in self.fx_list[-50:] if x['dt'] > self.bi_list[-1]['dt']]
            if self.bi_mode == "old":
                right_kn = [x for x in self.kline_new[-300:] if x['dt'] >= self.bi_list[-1]['dt']]
            elif self.bi_mode == 'new':
                right_kn = [x for x in self.kline_raw[-300:] if x['dt'] >= self.bi_list[-1]['dt']]
            else:
                raise ValueError

        for fx in right_fx:
            last_bi = self.bi_list[-1]
            bi = dict(fx)
            bi['bi'] = bi.pop('fx')
            if last_bi['fx_mark'] == fx['fx_mark']:
                if (last_bi['fx_mark'] == 'g' and last_bi['bi'] < bi['bi']) \
                        or (last_bi['fx_mark'] == 'd' and last_bi['bi'] > bi['bi']):
                    if self.verbose:
                        print("笔标记移动：from {} to {}".format(self.bi_list[-1], bi))
                    self.bi_list[-1] = bi
            else:
                kn_inside = [x for x in right_kn if last_bi['end_dt'] < x['dt'] < bi['start_dt']]
                if len(kn_inside) <= 0:
                    continue

                # 确保相邻两个顶底之间不存在包含关系
                if (last_bi['fx_mark'] == 'g' and bi['fx_low'] < last_bi['fx_low']
                    and bi['fx_high'] < last_bi['fx_high']) or \
                        (last_bi['fx_mark'] == 'd' and bi['fx_high'] > last_bi['fx_high']
                         and bi['fx_low'] > last_bi['fx_low']):
                    if self.verbose:
                        print("新增笔标记：{}".format(bi))
                    self.bi_list.append(bi)

    def _update_xd_list(self):
        """更新线段序列

        线段标记对象样例：
         {'dt': Timestamp('2020-07-09 00:00:00'),
          'fx_mark': 'g',
          'start_dt': Timestamp('2020-07-08 00:00:00'),
          'end_dt': Timestamp('2020-07-14 00:00:00'),
          'fx_high': 187.99,
          'fx_low': 163.12,
          'xd': 187.99}

         {'dt': Timestamp('2020-11-02 00:00:00'),
          'fx_mark': 'd',
          'start_dt': Timestamp('2020-10-29 00:00:00'),
          'end_dt': Timestamp('2020-11-03 00:00:00'),
          'fx_high': 142.38,
          'fx_low': 135.0,
          'xd': 135.0}
        """
        if len(self.bi_list) < 4:
            return

        self.xd_list = []
        if len(self.xd_list) == 0:
            for i in range(3):
                xd = dict(self.bi_list[i])
                xd['xd'] = xd.pop('bi')
                self.xd_list.append(xd)

        right_bi = [x for x in self.bi_list if x['dt'] >= self.xd_list[-1]['dt']]

        xd_p = get_potential_xd(right_bi)
        for xp in xd_p:
            xd = dict(xp)
            xd['xd'] = xd.pop('bi')
            last_xd = self.xd_list[-1]
            if last_xd['fx_mark'] == xd['fx_mark']:
                if (last_xd['fx_mark'] == 'd' and last_xd['xd'] > xd['xd']) \
                        or (last_xd['fx_mark'] == 'g' and last_xd['xd'] < xd['xd']):
                    if self.verbose:
                        print("更新线段标记：from {} to {}".format(last_xd, xd))
                    self.xd_list[-1] = xd
            else:
                if (last_xd['fx_mark'] == 'd' and last_xd['xd'] > xd['xd']) \
                        or (last_xd['fx_mark'] == 'g' and last_xd['xd'] < xd['xd']):
                    continue

                bi_inside = [x for x in right_bi if last_xd['dt'] <= x['dt'] <= xd['dt']]
                if len(bi_inside) < 4:
                    if self.verbose:
                        print("{} - {} 之间笔标记数量少于4，跳过".format(last_xd['dt'], xd['dt']))
                    continue
                else:
                    self.xd_list.append(xd)

    def update(self, k):
        """更新分析结果

        :param k: dict
            单根K线对象，样例如下
            {'symbol': '000001.SH',
             'dt': Timestamp('2020-07-16 15:00:00'),
             'open': 3356.11,
             'close': 3210.1,
             'high': 3373.53,
             'low': 3209.76,
             'vol': 486366915.0}
        """
        if self.verbose:
            print("=" * 100)
            print("输入新K线：{}".format(k))
        if not self.kline_raw or k['open'] != self.kline_raw[-1]['open']:
            self.kline_raw.append(k)
        else:
            if self.verbose:
                print("输入K线处于未完成状态，更新：replace {} with {}".format(self.kline_raw[-1], k))
            self.kline_raw[-1] = k

        if self.use_ta:
            self._update_ta()

        self._update_kline_new()
        self._update_fx_list()
        self._update_bi_list()

        if self.use_xd:
            self._update_xd_list()

        self.end_dt = self.kline_raw[-1]['dt']
        self.latest_price = self.kline_raw[-1]['close']

        if len(self.kline_raw) > self.max_count:
            last_dt = self.kline_raw[-self.max_count:][0]['dt']
            self.kline_raw = self.kline_raw[-self.max_count:]
            self.kline_new = self.kline_new[-self.max_count:]
            self.ma = [x for x in self.ma if x['dt'] > last_dt]
            self.macd = [x for x in self.macd if x['dt'] > last_dt]
            self.fx_list = [x for x in self.fx_list if x['dt'] > last_dt]
            self.bi_list = [x for x in self.bi_list if x['dt'] > last_dt]
            if self.use_xd:
                self.xd_list = [x for x in self.xd_list if x['dt'] > last_dt]

        if self.verbose:
            print("更新结束\n\n")

    def to_df(self, ma_params=(5, 20), use_macd=False, max_count=1000, mode="raw"):
        """整理成 df 输出

        :param ma_params: tuple of int
            均线系统参数
        :param use_macd: bool
        :param max_count: int
        :param mode: str
            使用K线类型， raw = 原始K线，new = 去除包含关系的K线
        :return: pd.DataFrame
        """
        if mode == "raw":
            bars = self.kline_raw[-max_count:]
        elif mode == "new":
            bars = self.kline_raw[-max_count:]
        else:
            raise ValueError

        fx_list = {x["dt"]: {"fx_mark": x["fx_mark"], "fx": x['fx']} for x in self.fx_list[-(max_count // 2):]}
        bi_list = {x["dt"]: {"fx_mark": x["fx_mark"], "bi": x['bi']} for x in self.bi_list[-(max_count // 4):]}
        xd_list = {x["dt"]: {"fx_mark": x["fx_mark"], "xd": x['xd']} for x in self.xd_list[-(max_count // 8):]}
        results = []
        for k in bars:
            k['fx_mark'], k['fx'], k['bi'], k['xd'] = "o", None, None, None
            fx_ = fx_list.get(k['dt'], None)
            bi_ = bi_list.get(k['dt'], None)
            xd_ = xd_list.get(k['dt'], None)
            if fx_:
                k['fx_mark'] = fx_["fx_mark"]
                k['fx'] = fx_["fx"]

            if bi_:
                k['bi'] = bi_["bi"]

            if xd_:
                k['xd'] = xd_["xd"]

            results.append(k)
        df = pd.DataFrame(results)
        for p in ma_params:
            df.loc[:, "ma{}".format(p)] = ta.SMA(df.close.values, p)
        if use_macd:
            diff, dea, macd = ta.MACD(df.close.values)
            df.loc[:, "diff"] = diff
            df.loc[:, "dea"] = diff
            df.loc[:, "macd"] = diff
        return df

    def to_image(self, file_image, mav=(5, 20, 120, 250), max_k_count=1000, dpi=50):
        """保存成图片

        :param file_image: str
            图片名称，支持 jpg/png/svg 格式，注意后缀
        :param mav: tuple of int
            均线系统参数
        :param max_k_count: int
            设定最大K线数量，这个值越大，生成的图片越长
        :param dpi: int
            图片分辨率
        :return:
        """
        ka_to_image(self, file_image=file_image, mav=mav, max_k_count=max_k_count, dpi=dpi)

    def calculate_macd_power(self, start_dt, end_dt, mode='bi', direction="up"):
        """用 MACD 计算走势段（start_dt ~ end_dt）的力度

        :param start_dt: datetime
            走势开始时间
        :param end_dt: datetime
            走势结束时间
        :param mode: str
            分段走势类型，默认值为 bi，可选值 ['bi', 'xd']，分别表示笔分段走势和线段分段走势
        :param direction: str
            线段分段走势计算力度需要指明方向，可选值 ['up', 'down']
        :return: float
            走势力度
        """
        if not self.use_ta:
            warnings.warn("没有进行辅助技术指标的计算，macd_power 返回 0")
            return 0

        fd_macd = [x for x in self.macd if end_dt >= x['dt'] >= start_dt]

        if mode == 'bi':
            power = sum([abs(x['macd']) for x in fd_macd])
        elif mode == 'xd':
            if direction == 'up':
                power = sum([abs(x['macd']) for x in fd_macd if x['macd'] > 0])
            elif direction == 'down':
                power = sum([abs(x['macd']) for x in fd_macd if x['macd'] < 0])
            else:
                raise ValueError
        else:
            raise ValueError
        return power

    def calculate_vol_power(self, start_dt, end_dt):
        """用 VOL 计算走势段（start_dt ~ end_dt）的力度

        :param start_dt: datetime
            走势开始时间
        :param end_dt: datetime
            走势结束时间
        :return: float
            走势力度
        """
        fd_vol = [x for x in self.kline_raw if end_dt >= x['dt'] >= start_dt]
        power = sum([x['vol'] for x in fd_vol])
        return int(power)

    def get_bi_fd(self, n=6):
        """假定本级别笔为次级别线段，进行走势分段

        fd 为 dict 对象，表示一段走势，可以是笔、线段，样例如下：

        fd = {
            "start_dt": "",
            "end_dt": "",
            "start_mark": p1,        # 笔开始标记
            "end_mark": p2,          # 笔结束标记
            "price_power": 0,        # 走势段价差
            "vol_power": 0,          # 成交量面积（柱子和）
            "direction": "up",
            "high": 0,
            "low": 0,
            "mode": "bi"
        }

        :param n:
        :return: list of dict
        """
        points = self.bi_list[-(n + 1):]
        assert len(points) == n + 1

        res = []
        for i in range(len(points) - 1):
            p1 = points[i]
            p2 = points[i + 1]
            direction = "up" if p1["bi"] < p2["bi"] else "down"
            # macd_power = self.calculate_macd_power(start_dt=p1['dt'], end_dt=p2['dt'], mode="bi", direction=direction)
            vol_power = self.calculate_vol_power(start_dt=p1['dt'], end_dt=p2['dt'])
            res.append({
                "start_dt": p1['dt'],
                "end_dt": p2['dt'],
                "start_mark": p1,
                "end_mark": p2,
                "price_power": abs(p1['bi'] - p2['bi']),
                "vol_power": vol_power,
                "direction": direction,
                "high": max(p1["bi"], p2["bi"]),
                "low": min(p1["bi"], p2["bi"]),
                "mode": "bi"
            })
        return res



# coding: utf-8
import os
import pandas as pd
import random
from czsc.utils import echarts_plot as plot
from czsc.analyze import KlineAnalyze

cur_path = os.path.split(os.path.realpath(__file__))[0]


def test_heat_map():
    data = [{"x": "{}hour".format(i), "y": "{}day".format(j), "heat": random.randint(0, 50)}
            for i in range(24) for j in range(7)]
    x_label = ["{}hour".format(i) for i in range(24)]
    y_label = ["{}day".format(i) for i in range(7)]
    hm = plot.heat_map(data, x_label=x_label, y_label=y_label)
    hm.render()


def test_kline_pro():
    file_kline = os.path.join(cur_path, "data/000001.SH_D.csv")
    kline = pd.read_csv(file_kline, encoding="utf-8")
    bars = kline.to_dict("records")
    ka = KlineAnalyze(bars)

    bs = []
    for x in ka.xd_list:
        if x['fx_mark'] == 'd':
            mark = "buy"
        else:
            mark = "sell"
        bs.append({"dt": x['dt'], "mark": mark, mark: x['xd']})

    chart = plot.kline_pro(ka.kline_raw, fx=ka.fx_list, bi=ka.bi_list, xd=ka.xd_list, bs=bs)
    chart.render()




# coding: utf-8
from typing import List


def cal_nbar_percentile(k: dict, kn: List[dict], n: int) -> float:
    """计算 N 周期区间百分位

    假设选股策略确认选股时间为 T，股价为 P，T+N 区间内的K线最高价为 MAX_P，最低价为 MIN_P，则
    N周期区间百分位 = (MAX_P - P) / (MAX_P - MIN_P) * 100;
    N周期区间百分位值域为(0, 100)，值越大，说明离区间最大值越远，对于做多交易而言有更高的安全性。

    :param k: dict
        信号出现时的 K 线，如：
        {'symbol': '000001.SH',
         'dt': '2007-04-05 15:00:00',
         'open': 3286.16,
         'close': 3319.14,
         'high': 3326.92,
         'low': 3259.63,
         'vol': 114528051.0}

    :param kn: List[dict]
        信号出现后的 N 根 K 线，如：
        [{'symbol': '000001.SH',
          'dt': '2007-04-06 15:00:00',
          'open': 3287.68,
          'close': 3323.59,
          'high': 3334.22,
          'low': 3273.86,
          'vol': 119644881.0},
         {'symbol': '000001.SH',
          'dt': '2007-04-09 15:00:00',
          'open': 3333.42,
          'close': 3398.95,
          'high': 3399.51,
          'low': 3333.26,
          'vol': 137314104.0}]

    :param n: int
        周期数 N

    :return: float
    """
    assert len(kn) == n, "计算 {} 周期区间百分位时传入的 kn 数量为 {}".format(n, len(kn))
    c = k['close']
    min_p = min([x['low'] for x in kn])
    max_p = max([x['high'] for x in kn])

    if max_p == min_p:
        return 0
    else:
        percentile = round((max_p-c) / (max_p-min_p) * 100, 2)
        return percentile

def cal_nbar_income(k: dict, kn: List[dict], n: int) -> float:
    """计算 N 周期区间收益

    假设选股策略确认选股时间为 T，股价为 P，第T+N 根K线收盘价为 C，则
    N周期绝对收益 = (C - P) / P * 100;
    N周期绝对收益值域为(-100, +inf)，值越大，说明绝对收益越大，对于做多交易而言有更高的收益率。

    :param k: dict
        信号出现时的 K 线，如：
        {'symbol': '000001.SH',
         'dt': '2007-04-05 15:00:00',
         'open': 3286.16,
         'close': 3319.14,
         'high': 3326.92,
         'low': 3259.63,
         'vol': 114528051.0}

    :param kn: List[dict]
        信号出现后的 N 根 K 线，如：
        [{'symbol': '000001.SH',
          'dt': '2007-04-06 15:00:00',
          'open': 3287.68,
          'close': 3323.59,
          'high': 3334.22,
          'low': 3273.86,
          'vol': 119644881.0},
         {'symbol': '000001.SH',
          'dt': '2007-04-09 15:00:00',
          'open': 3333.42,
          'close': 3398.95,
          'high': 3399.51,
          'low': 3333.26,
          'vol': 137314104.0}]

    :param n: int
        周期数 N

    :return: float
    """
    assert len(kn) == n, "计算 {} 周期区间收益时传入的 kn 数量为 {}".format(n, len(kn))
    c = k['close']
    last_c = kn[-1]['close']

    income = round((last_c - c) / c * 100, 2)
    return income














# coding: utf-8
from .analyze import KlineAnalyze


def check_jing(fd1, fd2, fd3, fd4, fd5) -> str:
    """检查最近5个分段走势是否构成井

    井的定义：
        12345，五段，是构造井的基本形态，形成井的位置肯定是5，而5出井的
        前提条件是对于向上5至少比3和1其中之一高，向下反过来; 并且，234
        构成一个中枢。

        井只有两类，大井和小井（以向上为例）：
        大井对应的形式是：12345向上，5最高3次之1最低，力度上1大于3，3大于5；
        小井对应的形式是：
            1：12345向上，3最高5次之1最低，力度上5的力度比1小，注意这时候
               不需要再考虑5和3的关系了，因为5比3低，所以不需要考虑力度。
            2：12345向上，5最高3次之1最低，力度上1大于5，5大于3。

        小井的构造，关键是满足5一定至少大于1、3中的一个。
        注意有一种情况不归为井：就是12345向上，1的力度最小，5的力度次之，3的力度最大此类不算井，
        因为345后面必然还有走势在67的时候才能再判断，个中道理各位好好体会。


    fd 为 dict 对象，表示一段走势，可以是笔、线段。

    假定最近一段走势为第N段；则 fd1 为第N-4段走势, fd2为第N-3段走势,
    fd3为第N-2段走势, fd4为第N-1段走势, fd5为第N段走势

    :param fd1: 第N-4段
    :param fd2: 第N-3段
    :param fd3: 第N-2段
    :param fd4: 第N-1段
    :param fd5: 第N段
    :return:
    """
    assert fd1['direction'] == fd3['direction'] == fd5['direction']
    assert fd2['direction'] == fd4['direction']
    direction = fd1['direction']

    zs_g = min(fd2['high'], fd3['high'], fd4['high'])
    zs_d = max(fd2['low'], fd3['low'], fd4['low'])

    jing = "other"

    if fd1['price_power'] < fd5['price_power'] < fd3['price_power'] \
            and fd1['vol_power'] < fd5['vol_power'] < fd3['vol_power']:
        # 1的力度最小，5的力度次之，3的力度最大，此类不算井
        return jing

    if zs_d < zs_g:  # 234有中枢的情况
        if direction == 'up' and fd5["high"] > min(fd3['high'], fd1['high']):
            # 大井: 12345向上，5最高3次之1最低，力度上1大于3，3大于5
            if fd5["high"] > fd3['high'] > fd1['high'] \
                    and fd5['price_power'] < fd3['price_power'] < fd1['price_power'] \
                    and fd5['vol_power'] < fd3['vol_power'] < fd1['vol_power']:
                jing = "向上大井"

            # 第一种小井: 12345向上，3最高5次之1最低，力度上5的力度比1小
            if fd1['high'] < fd5['high'] < fd3['high'] \
                    and fd5['price_power'] < fd1['price_power'] \
                    and fd5['vol_power'] < fd1['vol_power']:
                jing = "向上小井"

            # 第二种小井: 12345向上，5最高3次之1最低，力度上1大于5，5大于3
            if fd5["high"] > fd3['high'] > fd1['high'] \
                    and fd1['price_power'] > fd5['price_power'] > fd3['price_power'] \
                    and fd1['vol_power'] > fd5['vol_power'] > fd3['vol_power']:
                jing = "向上小井"

        if direction == 'down' and fd5["low"] < max(fd3['low'], fd1['low']):

            # 大井: 12345向下，5最低3次之1最高，力度上1大于3，3大于5
            if fd5['low'] < fd3['low'] < fd1['low'] \
                    and fd5['price_power'] < fd3['price_power'] < fd1['price_power'] \
                    and fd5['vol_power'] < fd3['vol_power'] < fd1['vol_power']:
                jing = "向下大井"

            # 第一种小井: 12345向下，3最低5次之1最高，力度上5的力度比1小
            if fd1["low"] > fd5['low'] > fd3['low'] \
                    and fd5['price_power'] < fd1['price_power'] \
                    and fd5['vol_power'] < fd1['vol_power']:
                jing = "向下小井"

            # 第二种小井: 12345向下，5最低3次之1最高，力度上1大于5，5大于3
            if fd5['low'] < fd3['low'] < fd1['low'] \
                    and fd1['price_power'] > fd5['price_power'] > fd3['price_power'] \
                    and fd1['vol_power'] > fd5['vol_power'] > fd3['vol_power']:
                jing = "向下小井"
    else:
        # 第三种小井
        if fd1['price_power'] > fd3['price_power'] > fd5['price_power'] \
                and fd1['vol_power'] > fd3['vol_power'] > fd5['vol_power']:
            if direction == 'up' and fd5["high"] > fd3['high'] > fd1['high']:
                # 12345类上涨趋势，力度依次降低
                jing = "向上小井"

            if direction == 'down' and fd5["low"] < fd3['low'] < fd1['low']:
                # 12345类下跌趋势，力度依次降低
                jing = "向下小井"

    return jing

def check_third_bs(fd1, fd2, fd3, fd4, fd5) -> str:
    """输入5段走势，判断是否存在第三类买卖点

    :param fd1: 第N-4段
    :param fd2: 第N-3段
    :param fd3: 第N-2段
    :param fd4: 第N-1段
    :param fd5: 第N段
    :return:
    """
    zs_d = max(fd1['low'], fd2['low'], fd3['low'])
    zs_g = min(fd1['high'], fd2['high'], fd3['high'])

    third_bs = "other"

    if fd5['high'] < zs_d < zs_g and fd4['low'] < min(fd1['low'], fd3['low']):
        third_bs = "三卖"

    if fd5['low'] > zs_g > zs_d and fd4['high'] > max(fd1['high'], fd3['high']):
        third_bs = "三买"

    return third_bs

def check_dynamic(fd1, fd3, fd5):
    """计算第N段走势的涨跌力度

    向上笔不创新高，向上笔新高盘背，向上笔新高无背
    向下笔不创新低，向下笔新低盘背，向下笔新低无背

    :param fd1: 第N-4段走势
    :param fd3: 第N-2段走势
    :param fd5: 第N段走势
    :return: str
    """
    if fd5['direction'] == "up":
        if fd5['high'] < fd3['high'] or fd5['high'] < fd1['high']:
            v = "向上笔不创新高"
        else:
            if fd5['price_power'] > fd3['price_power'] and fd5['price_power'] > fd1['price_power'] \
                    and fd5['vol_power'] > fd3['vol_power'] and fd5['vol_power'] > fd1['vol_power']:
                v = "向上笔新高无背"
            else:
                v = "向上笔新高盘背"
    elif fd5['direction'] == "down":
        if fd5['low'] > fd3['low'] or fd5['low'] > fd1['low']:
            v = "向下笔不创新低"
        else:
            if fd5['price_power'] > fd3['price_power'] and fd5['price_power'] > fd1['price_power'] \
                    and fd5['vol_power'] > fd3['vol_power'] and fd5['vol_power'] > fd1['vol_power']:
                v = "向下笔新低无背"
            else:
                v = "向下笔新低盘背"
    else:
        raise ValueError
    return v


class KlineSignals(KlineAnalyze):
    """适用于纯缠论逻辑推理的单级别信号计算"""

    def __init__(self, kline, name="本级别", bi_mode="new", max_count=300, use_xd=False, use_ta=False):
        super().__init__(kline, name, bi_mode, max_count, use_xd, use_ta,
                         ma_params=(5, 13, 21, 34, 55, 89, 144, 233), verbose=False)

    def get_signals(self):
        """获取单级别信号"""
        methods = [self.fx_signals, self.bi_signals, self.bd_signals]
        signals = dict()
        for method in methods:
            signals.update(method())
        return signals

    def fx_signals(self):
        """辅助判断的信号"""
        s = {
            "最近一个分型类型": "other",
            "最近三根无包含K线形态": "other",
            "最近一个底分型上边沿": 0,
            "最近一个顶分型下边沿": 0,
        }

        if len(self.fx_list) > 2:
            s['最近一个分型类型'] = self.fx_list[-1]['fx_mark']
            last_d = [x for x in self.fx_list[-4:] if x['fx_mark'] == 'd'][-1]
            last_g = [x for x in self.fx_list[-4:] if x['fx_mark'] == 'g'][-1]
            s['最近一个底分型上边沿'] = last_d['fx_high']
            s['最近一个顶分型下边沿'] = last_g['fx_low']

        def __tri_mark(x1, x2, x3):
            if x1 < x2 > x3:
                v = "g"
            elif x1 > x2 < x3:
                v = "d"
            elif x1 > x2 > x3:
                v = "down"
            elif x1 < x2 < x3:
                v = "up"
            else:
                v = 'other'
            return v

        if len(self.kline_new) > 6:
            k3 = self.kline_new[-3:]
            assert len(k3) == 3
            s['最近三根无包含K线形态'] = __tri_mark(k3[-3]['high'], k3[-2]['high'], k3[-1]['high'])

        return {"{}_{}".format(self.name, k): v for k, v in s.items()}

    def bi_signals(self):
        """笔相关信号"""
        s = {
            "五笔趋势类背驰": "other",  # other 表示默认值， up 表示向上五笔类趋势背驰， down 表示向下

            "第N笔涨跌力度": "other",  # other 表示默认值
            "第N笔向下新低": False,
            "第N笔向上新高": False,
            "第N笔结束标记的上边沿": 0,
            "第N笔结束标记的下边沿": 0,

            "第N-1笔涨跌力度": "other",
            "第N-2笔涨跌力度": "other",

            "第N笔第三买卖": "other",  # other 表示默认值
            "第N-2笔第三买卖": "other",

            "最近一个笔中枢ZD": 0,
            "最近一个笔中枢ZG": 0,

            "第N笔出井": "other",
            "第N-1笔出井": "other",
            "第N-2笔出井": "other",
        }

        if len(self.bi_list) > 12:
            bis = self.get_bi_fd(n=10)
            s['第N笔结束标记的上边沿'] = bis[-1]['end_mark']['fx_high']
            s['第N笔结束标记的下边沿'] = bis[-1]['end_mark']['fx_low']

            last_k = self.kline_new[-1]
            if bis[-1]['price_power'] < bis[-3]['price_power'] and bis[-1]['vol_power'] < bis[-3]['vol_power']:
                if bis[-5]['low'] > bis[-3]['low'] > bis[-1]['low'] \
                        and bis[-5]['high'] > bis[-3]['high'] > bis[-1]['high'] \
                        and bis[-1]['direction'] == 'down' and bis[-2]['high'] < bis[-4]['low']:
                    s['五笔趋势类背驰'] = 'down'

                if bis[-5]['low'] < bis[-3]['low'] < bis[-1]['low'] \
                        and bis[-5]['high'] < bis[-3]['high'] < bis[-1]['high'] \
                        and bis[-1]['direction'] == 'up' and bis[-2]['low'] > bis[-4]['high']:
                    s['五笔趋势类背驰'] = 'up'

            if bis[-1]['direction'] == 'down' and last_k['low'] < bis[-1]['low']:
                s['第N笔向下新低'] = True

            if bis[-1]['direction'] == 'up' and last_k['high'] > bis[-1]['high']:
                s['第N笔向上新高'] = True

            s['第N笔涨跌力度'] = check_dynamic(bis[-5], bis[-3], bis[-1])
            s['第N-1笔涨跌力度'] = check_dynamic(bis[-6], bis[-4], bis[-2])
            s['第N-2笔涨跌力度'] = check_dynamic(bis[-7], bis[-5], bis[-3])

            s['第N笔第三买卖'] = check_third_bs(bis[-5], bis[-4], bis[-3], bis[-2], bis[-1])
            s['第N-2笔第三买卖'] = check_third_bs(bis[-7], bis[-6], bis[-5], bis[-4], bis[-3])

            s['第N笔出井'] = check_jing(bis[-5], bis[-4], bis[-3], bis[-2], bis[-1])
            s['第N-1笔出井'] = check_jing(bis[-6], bis[-5], bis[-4], bis[-3], bis[-2])
            s['第N-2笔出井'] = check_jing(bis[-7], bis[-6], bis[-5], bis[-4], bis[-3])

            zd = max(bis[-5]['low'], bis[-4]['low'], bis[-3]['low'])
            zg = min(bis[-5]['high'], bis[-4]['high'], bis[-3]['high'])
            if zg < zd:
                zd = max(bis[-4]['low'], bis[-3]['low'], bis[-2]['low'])
                zg = min(bis[-4]['high'], bis[-3]['high'], bis[-2]['high'])

            s['最近一个笔中枢ZD'] = zd
            s['最近一个笔中枢ZG'] = zg
        return {"{}_{}".format(self.name, k): v for k, v in s.items()}

    def bd_signals(self):
        """由笔进行同级别分解的信号"""
        s = {
            "三笔回调构成第三买卖点": "other",
        }
        if len(self.bi_list) > 16 and self.bi_list[-1]['fx_mark'] == 'd':
            points = self.bi_list[-16:]
            fd1_points = [x['bi'] for x in points[0: 4]]
            fd2_points = [x['bi'] for x in points[3: 7]]
            fd3_points = [x['bi'] for x in points[6: 10]]
            fd4_points = [x['bi'] for x in points[9: 13]]
            fd5_points = [x['bi'] for x in points[12: 16]]
            fd1 = {"high": max(fd1_points), "low": min(fd1_points)}
            fd2 = {"high": max(fd2_points), "low": min(fd2_points)}
            fd3 = {"high": max(fd3_points), "low": min(fd3_points)}
            fd4 = {"high": max(fd4_points), "low": min(fd4_points)}
            fd5 = {"high": max(fd5_points), "low": min(fd5_points)}

            s['三笔回调构成第三买卖点'] = check_third_bs(fd1, fd2, fd3, fd4, fd5)
        return {"{}_{}".format(self.name, k): v for k, v in s.items()}


class MachineKlineSignals(KlineAnalyze):
    """适用于训练机器学习（多因子）模型的单级别信号计算"""

    def __init__(self, kline, name="本级别", bi_mode="new", max_count=300, use_xd=False, use_ta=False):
        super().__init__(kline, name, bi_mode, max_count, use_xd, use_ta,
                         ma_params=(5, 13, 21, 34, 55, 89, 144, 233), verbose=False)

    def get_signals(self):
        """获取单级别信号"""
        methods = [self.fx_signals, self.bi_signals, self.bd_signals]
        signals = dict()
        for method in methods:
            signals.update(method())
        return signals

    def fx_signals(self):
        """分型相关信号"""
        s = {
            "最近一个分型类型": "other",
            "最近三根无包含K线形态": "other",
        }
        if len(self.fx_list) > 2:
            s['最近一个分型类型'] = self.fx_list[-1]['fx_mark']

        def __tri_mark(x1, x2, x3):
            if x1 < x2 > x3:
                v = "g"
            elif x1 > x2 < x3:
                v = "d"
            elif x1 > x2 > x3:
                v = "down"
            elif x1 < x2 < x3:
                v = "up"
            else:
                v = 'other'
            return v

        if len(self.kline_new) > 6:
            k3 = self.kline_new[-3:]
            assert len(k3) == 3
            s['最近三根无包含K线形态'] = __tri_mark(k3[-3]['high'], k3[-2]['high'], k3[-1]['high'])

        return {"{}_{}".format(self.name, k): v for k, v in s.items()}

    def bi_signals(self):
        """笔相关信号"""
        s = {
            "当下笔方向": "other",  # other 表示默认值
            "五笔趋势类背驰": "other",  # other 表示默认值， up 表示向上五笔类趋势背驰， down 表示向下

            "第N笔涨跌力度": "other",  # other 表示默认值
            "第N笔第三买卖": "other",  # other 表示默认值

            "第N-1笔涨跌力度": "other",
            "第N-1笔第三买卖": "other",

            "第N-2笔涨跌力度": "other",
            "第N-2笔第三买卖": "other",

            "第N-3笔涨跌力度": "other",
            "第N-3笔第三买卖": "other",

            "最近一个笔中枢ZD": 0,
            "最近一个笔中枢ZG": 0,

            "第N-1笔创近6笔新高": False,
            "第N-1笔创近6笔新低": False,

            # 止损信号
            "当下笔向下新低": -1,  # -1 表示默认值， 0 表示 False， 1 表示 True
            "当下笔向上新高": -1,  # -1 表示默认值， 0 表示 False， 1 表示 True
        }

        if len(self.bi_list) > 12:
            bis = self.get_bi_fd(n=10)
            s['当下笔方向'] = bis[-1]['direction']
            last_k = self.kline_new[-1]
            if bis[-1]['price_power'] < bis[-3]['price_power'] and bis[-1]['vol_power'] < bis[-3]['vol_power']:
                if bis[-5]['low'] > bis[-3]['low'] > bis[-1]['low'] \
                        and bis[-5]['high'] > bis[-3]['high'] > bis[-1]['high'] \
                        and bis[-1]['direction'] == 'down' and bis[-2]['high'] < bis[-4]['low']:
                    s['五笔趋势类背驰'] = 'down'

                if bis[-5]['low'] < bis[-3]['low'] < bis[-1]['low'] \
                        and bis[-5]['high'] < bis[-3]['high'] < bis[-1]['high'] \
                        and bis[-1]['direction'] == 'up' and bis[-2]['low'] > bis[-4]['high']:
                    s['五笔趋势类背驰'] = 'up'

            if bis[-1]['direction'] == 'down':
                s['当下笔向下新低'] = 1 if last_k['low'] < bis[-1]['low'] else 0

            if bis[-1]['direction'] == 'up':
                s['当下笔向上新高'] = 1 if last_k['high'] > bis[-1]['high'] else 0

            s['第N-1笔创近6笔新高'] = max([x['high'] for x in bis[-7:-1]]) == bis[-2]['high'] and bis[-2]['direction'] == 'up'
            s['第N-1笔创近6笔新低'] = min([x['low'] for x in bis[-7:-1]]) == bis[-2]['low'] and bis[-2]['direction'] == 'down'

            s['第N笔涨跌力度'] = check_dynamic(bis[-5], bis[-3], bis[-1])
            s['第N-1笔涨跌力度'] = check_dynamic(bis[-6], bis[-4], bis[-2])
            s['第N-2笔涨跌力度'] = check_dynamic(bis[-7], bis[-5], bis[-3])
            s['第N-3笔涨跌力度'] = check_dynamic(bis[-8], bis[-6], bis[-4])

            s['第N笔第三买卖'] = check_third_bs(bis[-5], bis[-4], bis[-3], bis[-2], bis[-1])
            s['第N-1笔第三买卖'] = check_third_bs(bis[-6], bis[-5], bis[-4], bis[-3], bis[-2])
            s['第N-2笔第三买卖'] = check_third_bs(bis[-7], bis[-6], bis[-5], bis[-4], bis[-3])
            s['第N-3笔第三买卖'] = check_third_bs(bis[-8], bis[-7], bis[-6], bis[-5], bis[-4])

            zd = max(bis[-5]['low'], bis[-4]['low'], bis[-3]['low'])
            zg = min(bis[-5]['high'], bis[-4]['high'], bis[-3]['high'])
            if zg < zd:
                zd = max(bis[-4]['low'], bis[-3]['low'], bis[-2]['low'])
                zg = min(bis[-4]['high'], bis[-3]['high'], bis[-2]['high'])

            s['最近一个笔中枢ZD'] = zd
            s['最近一个笔中枢ZG'] = zg
        return {"{}_{}".format(self.name, k): v for k, v in s.items()}

    def bd_signals(self):
        """由笔进行同级别分解的信号"""
        s = {
            "三笔回调构成第三买卖点": "other",
        }
        if len(self.bi_list) > 16 and self.bi_list[-1]['fx_mark'] == 'd':
            points = self.bi_list[-16:]
            fd1_points = [x['bi'] for x in points[0: 4]]
            fd2_points = [x['bi'] for x in points[3: 7]]
            fd3_points = [x['bi'] for x in points[6: 10]]
            fd4_points = [x['bi'] for x in points[9: 13]]
            fd5_points = [x['bi'] for x in points[12: 16]]
            fd1 = {"high": max(fd1_points), "low": min(fd1_points)}
            fd2 = {"high": max(fd2_points), "low": min(fd2_points)}
            fd3 = {"high": max(fd3_points), "low": min(fd3_points)}
            fd4 = {"high": max(fd4_points), "low": min(fd4_points)}
            fd5 = {"high": max(fd5_points), "low": min(fd5_points)}

            s['三笔回调构成第三买卖点'] = check_third_bs(fd1, fd2, fd3, fd4, fd5)
        return {"{}_{}".format(self.name, k): v for k, v in s.items()}



# coding: utf-8

import pandas as pd
import mplfinance as mpf
import matplotlib as mpl
import matplotlib.pyplot as plt
from .echarts_plot import kline_pro


def ka_to_image(ka, file_image, mav=(5, 20, 120, 250), max_k_count=1000, dpi=50):
    """绘制 ka，保存到 file_image"""
    df = ka.to_df(use_macd=True, ma_params=(5, 20,), max_count=max_k_count)
    df.rename({"open": "Open", "close": "Close", "high": "High",
               "low": "Low", "vol": "Volume"}, axis=1, inplace=True)
    df.index = pd.to_datetime(df['dt'])
    df = df.tail(max_k_count)
    kwargs = dict(type='candle', mav=mav, volume=True)

    bi_xd = [
        [(x['dt'], x['bi']) for _, x in df.iterrows() if x['bi'] > 0],
        [(x['dt'], x['xd']) for _, x in df.iterrows() if x['xd'] > 0]
    ]

    mc = mpf.make_marketcolors(
        up='red',
        down='green',
        edge='i',
        wick='i',
        volume='in',
        inherit=True)

    s = mpf.make_mpf_style(
        gridaxis='both',
        gridstyle='-.',
        y_on_right=False,
        marketcolors=mc)

    mpl.rcParams['font.sans-serif'] = ['KaiTi']
    mpl.rcParams['font.serif'] = ['KaiTi']
    mpl.rcParams['font.size'] = 48
    mpl.rcParams['axes.unicode_minus'] = False
    mpl.rcParams['lines.linewidth'] = 1.0

    title = '%s@%s（%s - %s）' % (ka.symbol, ka.name, df.index[0].__str__(), df.index[-1].__str__())
    fig, axes = mpf.plot(df, columns=['Open', 'High', 'Low', 'Close', 'Volume'], style=s,
                         title=title, ylabel='K线', ylabel_lower='成交量', **kwargs,
                         alines=dict(alines=bi_xd, colors=['r', 'g'], linewidths=8, alpha=0.35),
                         returnfig=True)

    w = len(df) * 0.15
    fig.set_size_inches(w, 30)
    ax = plt.gca()
    ax.set_xbound(-1, len(df) + 1)
    fig.savefig(fname=file_image, dpi=dpi, bbox_inches='tight')
    plt.close()


def ka_to_echarts(ka, width: str = "1500px", height: str = '800px'):
    """用 pyecharts 绘制分析结果

    :param ka: KlineAnalyze
    :param width: str
    :param height: str
    :return:
    """
    symbol = ka.kline_raw[0]['symbol']
    title = "{} - {}".format(symbol, ka.name)
    chart = kline_pro(ka.kline_raw, fx=ka.fx_list, title=title,
                      bi=ka.bi_list, xd=ka.xd_list, width=width, height=height)
    return chart



# coding: utf-8
import os
import pandas as pd
from czsc.analyze import KlineAnalyze, find_zs

cur_path = os.path.split(os.path.realpath(__file__))[0]

def test_ka_update():
    file_kline = os.path.join(cur_path, "data/000001.SH_D.csv")
    kline = pd.read_csv(file_kline, encoding="utf-8")
    kline.loc[:, "dt"] = pd.to_datetime(kline.dt)
    kline1 = kline.iloc[:2000]
    kline2 = kline.iloc[2000:]

    ka1 = KlineAnalyze(kline, name="日线", max_count=1000, use_xd=True, verbose=False)
    ka2 = KlineAnalyze(kline1, name="日线", max_count=1000, use_xd=True, verbose=False)

    for _, row in kline2.iterrows():
        ka2.update(row.to_dict())

    assert ka1.kline_new[-1]['dt'] == ka2.kline_new[-1]['dt']
    assert ka1.fx_list[-1]['dt'] == ka2.fx_list[-1]['dt']
    assert ka1.bi_list[-1]['dt'] == ka2.bi_list[-1]['dt']
    assert ka1.xd_list[-1]['dt'] == ka2.xd_list[-1]['dt']

    ka3 = KlineAnalyze(kline, name="日线", max_count=1000, use_xd=False, verbose=False)
    assert ka3.kline_new[-1]['dt'] == ka2.kline_new[-1]['dt']
    assert ka3.fx_list[-1]['dt'] == ka2.fx_list[-1]['dt']
    assert ka3.bi_list[-1]['dt'] == ka2.bi_list[-1]['dt']
    assert not ka3.xd_list

def test_calculate_power():
    file_kline = os.path.join(cur_path, "data/000001.SH_D.csv")
    kline = pd.read_csv(file_kline, encoding="utf-8")
    kline.loc[:, "dt"] = pd.to_datetime(kline.dt)
    ka = KlineAnalyze(kline, name="日线", max_count=1000, use_xd=True, verbose=False)

    # 测试 macd 力度
    last_xd_power = ka.calculate_macd_power(start_dt=ka.xd_list[-2]['dt'], end_dt=ka.xd_list[-1]['dt'],
                                            mode='xd', direction="up" if ka.xd_list[-1]['fx_mark'] == 'g' else "down")

    last_bi_power = ka.calculate_macd_power(start_dt=ka.bi_list[-2]['dt'], end_dt=ka.bi_list[-1]['dt'], mode='bi')

    assert int(last_xd_power) == 389
    assert int(last_bi_power) == 300

    # 测试 vol 力度
    last_xd_power = ka.calculate_vol_power(start_dt=ka.xd_list[-2]['dt'], end_dt=ka.xd_list[-1]['dt'])
    last_bi_power = ka.calculate_vol_power(start_dt=ka.bi_list[-2]['dt'], end_dt=ka.bi_list[-1]['dt'])

    assert int(last_xd_power) == 13329239053
    assert int(last_bi_power) == 9291793337

def test_kline_analyze():
    file_kline = os.path.join(cur_path, "data/000001.SH_D.csv")
    kline = pd.read_csv(file_kline, encoding="utf-8")
    kline.loc[:, "dt"] = pd.to_datetime(kline.dt)
    ka = KlineAnalyze(kline, name="日线", max_count=1000, use_xd=True, verbose=False)

    # 测试绘图
    file_img = "kline.png"
    ka.to_image(file_img, max_k_count=5000)
    assert os.path.exists(file_img)

    # 测试分型识别结果
    assert ka.fx_list[-1]['fx_mark'] == 'g'
    assert ka.fx_list[-5]['fx_mark'] == 'g'

    # 测试笔识别结果
    assert ka.bi_list[-1]['fx_mark'] == 'g'
    assert ka.bi_list[-4]['fx_mark'] == 'd'

    # 测试线段识别结果
    assert ka.xd_list[-2]['fx_mark'] == 'g'
    assert ka.xd_list[-3]['fx_mark'] == 'd'

    # 测试增量更新
    for x in [2890, 2910, 2783, 3120]:
        k = dict(ka.kline_raw[-1])
        k['close'] = x
        ka.update(k)
        assert ka.kline_raw[-1]['close'] == x

def test_update_ta():
    file_kline = os.path.join(cur_path, "data/000001.SH_D.csv")
    kline = pd.read_csv(file_kline, encoding="utf-8")
    kline.loc[:, "dt"] = pd.to_datetime(kline.dt)
    ka = KlineAnalyze(kline, name="日线", max_count=1000, use_xd=True, verbose=False, use_ta=True)

    ma_x1 = dict(ka.ma[-1])
    macd_x1 = dict(ka.macd[-1])
    ka.update(kline.iloc[-1].to_dict())
    ma_x2 = dict(ka.ma[-1])
    macd_x2 = dict(ka.macd[-1])
    assert ma_x1['dt'] == ma_x2['dt']
    assert [round(x, 2) for x in ma_x1.values() if isinstance(x, float)] == \
           [round(x, 2) for x in ma_x2.values() if isinstance(x, float)]

    assert macd_x1['dt'] == macd_x2['dt']
    assert [round(x, 2) for x in macd_x1.values() if isinstance(x, float)] == \
           [round(x, 2) for x in macd_x2.values() if isinstance(x, float)]


def test_find_zs():
    # 造数测试
    points = [
        {"dt": 0, "fx_mark": "d", "xd": 8},
        {"dt": 1, "fx_mark": "g", "xd": 10},
        {"dt": 2, "fx_mark": "d", "xd": 9},
        {"dt": 3, "fx_mark": "g", "xd": 11},
        {"dt": 4, "fx_mark": "d", "xd": 10.5},
        {"dt": 5, "fx_mark": "g", "xd": 12},
        {"dt": 6, "fx_mark": "d", "xd": 11.1},

        {"dt": 7, "fx_mark": "g", "xd": 14},
        {"dt": 8, "fx_mark": "d", "xd": 13},
        {"dt": 9, "fx_mark": "g", "xd": 13.8},
        {"dt": 10, "fx_mark": "d", "xd": 12.9},
        {"dt": 11, "fx_mark": "g", "xd": 14.5},
        {"dt": 12, "fx_mark": "d", "xd": 13.2},
        {"dt": 13, "fx_mark": "g", "xd": 15},
        {"dt": 14, "fx_mark": "d", "xd": 14.3},

        {"dt": 15, "fx_mark": "g", "xd": 16.2},
        {"dt": 16, "fx_mark": "d", "xd": 15.3},
        {"dt": 17, "fx_mark": "g", "xd": 17.6},
        {"dt": 18, "fx_mark": "d", "xd": 15.9},
        {"dt": 19, "fx_mark": "g", "xd": 18.2},
        {"dt": 20, "fx_mark": "d", "xd": 16.8},
        {"dt": 21, "fx_mark": "g", "xd": 17.8},
        {"dt": 22, "fx_mark": "d", "xd": 16.9},
        {"dt": 23, "fx_mark": "g", "xd": 18.1},
    ]
    zss = find_zs(points[:8])
    assert len(zss) == 1

    zss = find_zs(points[:15])
    assert len(zss) == 2

    zss = find_zs(points)
    assert len(zss) == 3 and zss[0]['ZG'] < zss[1]['ZD'] and zss[1]['ZG'] < zss[2]['ZD']

    # 获取用于比较趋势背驰的两端
    fd1 = [x for x in points if x['dt'] >= zss[2]['end_point']['dt']]
    fd2 = [x for x in points if zss[2]['start_point']['dt'] > x['dt'] >= zss[1]['end_point']['dt']]
    fd3 = [x for x in points if zss[1]['start_point']['dt'] > x['dt'] >= zss[0]['end_point']['dt']]
    fd4 = [x for x in points if x['dt'] <= zss[0]['start_point']['dt']]
    assert fd1[0]['fx_mark'] == fd2[0]['fx_mark'] == fd3[0]['fx_mark'] == fd4[0]['fx_mark'] == 'd'
    assert fd1[-1]['fx_mark'] == fd2[-1]['fx_mark'] == fd3[-1]['fx_mark'] == fd4[-1]['fx_mark'] == 'g'






# coding: utf-8
import pandas as pd
import tushare as ts
from datetime import datetime, timedelta


pro = ts.pro_api()


def set_token(token):
    """在同一台机器上只需要调用 set_token 一次就可以

    :param token: str
        tushare.pro 的 token，如果没有，请到这里注册：https://tushare.pro/register?reg=7
    :return: None
    """
    ts.set_token(token)


def get_token():
    """获取调用凭证"""
    return ts.get_token()


def get_concepts():
    """获取概念列表

    https://dataapi.joinquant.com/docs#get_concepts---%E8%8E%B7%E5%8F%96%E6%A6%82%E5%BF%B5%E5%88%97%E8%A1%A8

    :return: df
    """
    return pro.concept(src='ts')

def get_concept_stocks(symbol, date=None):
    """获取概念成份股

    https://tushare.pro/document/2?doc_id=126

    :param symbol: str
        如 GN036
    :param date: str or datetime
        日期，如 2020-08-08
    :return: list

    examples:
    -------
    >>> symbols1 = get_concept_stocks("GN036", date="2020-07-08")
    >>> symbols2 = get_concept_stocks("GN036", date=datetime.now())
    """
    del date
    df = pro.concept_detail(id=symbol, fields='ts_code')
    return list(set([x + "-E" for x in df.ts_code]))

def get_index_stocks(symbol, date=None):
    """获取指数成份股

    https://dataapi.joinquant.com/docs#get_index_stocks---%E8%8E%B7%E5%8F%96%E6%8C%87%E6%95%B0%E6%88%90%E4%BB%BD%E8%82%A1

    :param symbol: str
        如 399300.SZ
    :param date: str or datetime
        日期，如 2020-08-08
    :return: list

    examples:
    -------
    >>> symbols1 = get_index_stocks("000300.XSHG", date="2020-07-08")
    >>> symbols2 = get_index_stocks("000300.XSHG", date=datetime.now())
    """
    if not date:
        date = datetime.now()

    if isinstance(date, str):
        date = pd.to_datetime(date)

    start_date = date - timedelta(days=250)
    end_date = date

    df = pro.index_weight(index_code=symbol, start_date=start_date.strftime("%Y%m%d"),
                          end_date=end_date.strftime("%Y%m%d"))
    return list(set([x + "-E" for x in df.con_code]))


def _get_start_date(end_date, freq):
    if isinstance(end_date, str):
        end_date = pd.to_datetime(end_date)

    if freq == '1min':
        start_date = end_date - timedelta(days=30)
    elif freq == '5min':
        start_date = end_date - timedelta(days=70)
    elif freq == '30min':
        start_date = end_date - timedelta(days=500)
    elif freq == 'D':
        start_date = end_date - timedelta(weeks=500)
    elif freq == 'W':
        start_date = end_date - timedelta(weeks=1000)
    elif freq == 'M':
        start_date = end_date - timedelta(weeks=2000)
    else:
        raise ValueError("'freq' value error, current value is %s, "
                         "optional valid values are ['1min', '5min', '30min', "
                         "'D', 'W']" % freq)
    return start_date

def get_kline(symbol,  end_date, freq, start_date=None, count=None):
    """获取K线数据

    :param symbol: str
        Tushare 标的代码 + Tushare asset 代码，如 000001.SH-I
    :param start_date: datetime
        截止日期
    :param end_date: datetime
        截止日期
    :param freq: str
        K线级别，可选值 ['1min', '5min', '30min', '60min', 'D', 'W', "M"]
    :param count: int
        K线数量，最大值为 5000
    :return: pd.DataFrame

    >>> start_date = datetime.strptime("20200701", "%Y%m%d")
    >>> end_date = datetime.strptime("20200719", "%Y%m%d")
    >>> df1 = get_kline(symbol="000001.SH-I", start_date=start_date, end_date=end_date, freq="1min")
    >>> df2 = get_kline(symbol="000001.SH-I", end_date=end_date, freq="1min", count=1000)
    """
    ts_code, asset = symbol.split("-")
    if count:
        start_date = _get_start_date(end_date, freq)
        start_date = start_date.date().__str__().replace("-", "")

        if isinstance(end_date, str):
            end_date = pd.to_datetime(end_date)

        end_date = end_date + timedelta(days=1)
        end_date = end_date.date().__str__().replace("-", "")

    if isinstance(end_date, datetime):
        end_date = end_date.date().__str__().replace("-", "")

    if isinstance(start_date, datetime):
        start_date = start_date.date().__str__().replace("-", "")

    df = ts.pro_bar(ts_code=ts_code, freq=freq, start_date=start_date, end_date=end_date,
                    adj='qfq', asset=asset)

    # 统一 k 线数据格式为 6 列，分别是 ["symbol", "dt", "open", "close", "high", "low", "vr"]
    if "min" in freq:
        df.rename(columns={'ts_code': "symbol", "trade_time": "dt"}, inplace=True)
    else:
        df.rename(columns={'ts_code': "symbol", "trade_date": "dt"}, inplace=True)

    df.drop_duplicates(subset='dt', keep='first', inplace=True)
    df.sort_values('dt', inplace=True)
    df['dt'] = df.dt.apply(str)
    if freq.endswith("min"):
        # 清理 9:30 的空数据
        df['not_start'] = df.dt.apply(lambda x: not x.endswith("09:30:00"))
        df = df[df['not_start']]
    if count:
        df = df.tail(count)

    df.reset_index(drop=True, inplace=True)
    df.loc[:, "dt"] = pd.to_datetime(df['dt'])

    k = df[['symbol', 'dt', 'open', 'close', 'high', 'low', 'vol']]

    for col in ['open', 'close', 'high', 'low']:
        k[col] = k[col].apply(round, args=(2,))

    return k


def download_kline(symbol, freq, start_date, end_date, delta, save=True):
    """下载K线数据

    :param save:
    :param symbol:
    :param end_date:
    :param freq:
    :param start_date:
    :param delta:
    :return:

    >>> start_date = datetime.strptime("20200101", "%Y%m%d")
    >>> end_date = datetime.strptime("20200719", "%Y%m%d")
    >>> df = download_kline("000001.SH-I", "1min", start_date, end_date, delta=timedelta(days=10), save=False)
    """
    data = []
    end_dt = start_date + delta
    print("开始下载数据：{} - {} - {}".format(symbol, start_date, end_date))
    df_ = get_kline(symbol, start_date=start_date, end_date=end_dt, freq=freq)
    if not df_.empty:
        data.append(df_)

    while end_dt < end_date:
        df_ = get_kline(symbol, start_date=start_date, end_date=end_dt, freq=freq)
        if not df_.empty:
            data.append(df_)
        start_date = end_dt
        end_dt += delta
        print("当前下载进度：{} - {} - {}".format(symbol, start_date, end_dt))

    df = pd.concat(data, ignore_index=True)
    print("{} 去重前K线数量为 {}".format(symbol, len(df)))
    df.drop_duplicates(['dt'], inplace=True)
    df.sort_values('dt', ascending=True, inplace=True)
    df.reset_index(drop=True, inplace=True)
    print("{} 去重后K线数量为 {}".format(symbol, len(df)))

    if save:
        df.to_csv(f"{symbol}_{freq}_{start_date.date()}_{end_date.date()}.csv", index=False, encoding="utf-8")
    else:
        return df



# coding: utf-8

from .plot import ka_to_echarts, ka_to_image
from .echarts_plot import kline_pro, heat_map
from .kline_generator import KlineGeneratorBy1Min, KlineGeneratorByTick
from .ta import KDJ, MACD, EMA, SMA

